--hybridContrl version 0.0.4alpha
--Final Edit 15点47分2023年12月24日
--by NZZ

local M = {}

local proxyEngine = nil
local gearbox = nil
local motors = nil 

local controlLogicModule = nil
local mode = nil
local isCrawl = 0
local ifMotorOn = nil
local motorRatio1 = nil
local motorRatio2 = nil

local detN = nil

local startVelocity = nil
local connectVelocity = nil

local enableModes = {}

local function reduceRegen()
    for _, v in ipairs(motors) do
        if v.maxWantedRegenTorque > 0 then
            v.maxWantedRegenTorque = v.maxWantedRegenTorque - v.originalRegenTorque * 0.2
            gui.message({ txt = "Energy Recovery Reduced" }, 5, "", "")
        elseif v.maxWantedRegenTorque == 0 then
            gui.message({ txt = "Energy Recovery Min" }, 5, "", "")
        elseif v.maxWantedRegenTorque < 0 then
            v.maxWantedRegenTorque = 0
            gui.message({ txt = "Energy Recovery Min" }, 5, "", "")
        end
    end

end

local function enhanceRegen()
    for _, v in ipairs(motors) do
        if v.maxWantedRegenTorque < v.originalRegenTorque then
            v.maxWantedRegenTorque = v.maxWantedRegenTorque + v.originalRegenTorque * 0.2
            gui.message({ txt = "Energy Recovery Enhanced" }, 5, "", "")
        elseif v.maxWantedRegenTorque == v.originalRegenTorque then
            gui.message({ txt = "Energy Recovery Max" }, 5, "", "")
        elseif v.maxWantedRegenTorque > v.originalRegenTorque then
            v.maxWantedRegenTorque = v.originalRegenTorque
            gui.message({ txt = "Energy Recovery Max" }, 5, "", "")
        end
    end
end

local function getGear()
    if ifMotorOn == true then
        if gearbox.mode then
            electrics.values.gearName = gearbox.mode
            if gearbox.mode == "drive" then -- D gear , S gear , R gear or M gear
                electrics.values.motorDirection = gearbox.gearIndex
            elseif gearbox.mode == "reverse" then -- CVT R gear
                electrics.values.motorDirection = -1
            elseif gearbox.mode == "neutral" then -- N gear
                electrics.values.motorDirection = 0
            elseif gearbox.mode == "park" then -- P gear
                electrics.values.motorDirection = 0
            end
        else
            electrics.values.motorDirection = gearbox.gearIndex
        end
    elseif ifMotorOn == false then
        electrics.values.motorDirection = 0
    end

end

local function setMode(mode)
    for _, u in ipairs(enableModes) do
        if mode == u then
            electrics.values.hybridMode = mode

            if mode == "hybrid" then
                gui.message({ txt = "Switch to hybrid mode" }, 5, "", "")
                for _, v in ipairs(motors) do
                    v:setmotorRatio(motorRatio1)
                    v:setMode("connected")
                    ifMotorOn = true
                    if electrics.values.ignitionLevel == 2 and electrics.values.engineRunning == 0 then
                        proxyEngine:activateStarter()
                    end
                end
            elseif mode == "fuel" then
                gui.message({ txt = "Switch to fuel mode" }, 5, "", "")
                for _, v in ipairs(motors) do
                    v:setmotorRatio(0)
                    v:setMode("connected")
                    ifMotorOn = true
                    if electrics.values.ignitionLevel == 2 and electrics.values.engineRunning == 0 then
                        proxyEngine:activateStarter()
                    end
                end
            elseif mode == "electric" then
                gui.message({ txt = "Switch to electric mode" }, 5, "", "")
                for _, v in ipairs(motors) do
                    v:setmotorRatio(motorRatio2)
                    v:setMode("disconnected")
                    ifMotorOn = true
                    proxyEngine:setIgnition(0)
                end
            elseif mode == "auto" then
                gui.message({ txt = "Switch to auto mode" }, 5, "", "")
                for _, v in ipairs(motors) do
                    v:setmotorRatio(motorRatio2)
                    v:setMode("disconnected")
                    ifMotorOn = true
                    proxyEngine:setIgnition(0)
                end
            end

        end
    end
end

local function updateGFX(dt)
    --log("", "hybrid", "hybrid" .. 3)
    getGear()

    if electrics.values.hybridMode == "electric" then
        proxyEngine:setIgnition(0)
    end

    --auto mode
    if electrics.values.hybridMode == "auto" then
        if electrics.values.airspeed < startVelocity - 5 and detN ~= 1 then
            for _, v in ipairs(motors) do
                v:setmotorRatio(motorRatio2)
                v:setMode("disconnected")
                ifMotorOn = true
                proxyEngine:setIgnition(0)
            end
            detN = 1
        elseif electrics.values.airspeed >= startVelocity and electrics.values.airspeed < connectVelocity and detN ~= 2 then
            detN = 2
        elseif electrics.values.airspeed >= connectVelocity and detN ~= 3 then
            for _, v in ipairs(motors) do
                v:setmotorRatio(motorRatio1)
                v:setMode("connected")
                ifMotorOn = true
                if electrics.values.ignitionLevel == 2 and electrics.values.engineRunning == 0 then
                    proxyEngine:activateStarter()
                end
            end
            detN = 3
        end
    
        if electrics.values.airspeed < startVelocity - 5 then
            if electrics.values.engineRunning == 1 then  
                proxyEngine:setIgnition(0)
            end
        elseif electrics.values.airspeed >= startVelocity and electrics.values.airspeed < connectVelocity then
            if electrics.values.engineRunning == 0 then
                proxyEngine:activateStarter()
            end
        end
    end

    if electrics.values.hybridMode ~= "auto" then
        detN = 0
    end
    --auto mode
end

local function init(jbeamData)

    enableModes = jbeamData.enableModes or {"hybrid", "fuel", "electric", "auto"}
    --auto      自动选择
    --reev      增程模式
    --electric  电机直驱
    --fuel      燃油直驱
    --hybrid    混合驱动
    
    motorRatio1 = jbeamData.motorRatio1 or 1
    motorRatio2 = jbeamData.motorRatio2 or 1
    startVelocity = jbeamData.startVelocity or 10
    connectVelocity = jbeamData.connectVelocity or 11

    proxyEngine = powertrain.getDevice("mainEngine")
    gearbox = powertrain.getDevice("gearbox")
    
    motors = {}
    local motorNames = jbeamData.motorNames or {"mainMotor"}
    for _, v in ipairs(motorNames) do
        local motor = powertrain.getDevice(v)
        if motor then
            table.insert(motors, motor)
            motor.originalRegenTorque = motor.maxWantedRegenTorque
        end
    end
    if jbeamData.defaultMode then
        setMode(jbeamData.defaultMode)
    else
        setMode("hybrid")
    end
    
end

local function new()
end

local function onInit()
end

local function onReset(jbeamData)
    if jbeamData.defaultMode then
        setMode(jbeamData.defaultMode)
    else
        setMode("hybrid")
    end
end

M.setMode = setMode

M.reduceRegen = reduceRegen
M.enhanceRegen = enhanceRegen

M.init = init
M.onInit = onInit
M.onReset = onReset

M.new = new
M.updateGFX = updateGFX


return M